package crystal.ast;

import crystal.general.*;
import crystal.emitc.*;

/** This class incorporates the logic about those of gcc's special __builtin_
 * functions we support, and helps with tasks such as computation of their
 * types and output back into C */

public class Builtins {
	private static final String BI_VA_ARG_NAME   = "__builtin_va_arg";
	private static final String BI_OFFSETOF_NAME = "__builtin_offsetof";
	private static final String BI_VA_START_NAME = "__builtin_va_start";
	private static final String BI_VA_END_NAME   = "__builtin_va_end";

	private static final String[] builtins = 
	{
		BI_VA_ARG_NAME, 
		BI_OFFSETOF_NAME,
		BI_VA_START_NAME,
		BI_VA_END_NAME
	};
	
	/* for the parser */
	public static Symbol BI_VA_ARG;
	public static Symbol BI_OFFSETOF;
	public static Symbol BI_VA_START;
	public static Symbol BI_VA_END;

	/** Call this function to enter the builtins into the global symbol table,
	 * and initialize our own hash of relevant builtins */
	public static void registerBuiltins() {
		for (String id: builtins)
			registerBuiltin(id);

		HashSymtab global = Symtab.getGlobalScope();
		BI_VA_ARG   = global.lookupVFTSelf(BI_VA_ARG_NAME);
		BI_OFFSETOF = global.lookupVFTSelf(BI_OFFSETOF_NAME);
		BI_VA_START = global.lookupVFTSelf(BI_VA_START_NAME);
		BI_VA_END   = global.lookupVFTSelf(BI_VA_END_NAME);
	}

	private static void registerBuiltin(String id) {
		/* For simplicity, we enter as void-to-void functions since we 
		 * can't produce a sane type for these pseudo-functions, anyway. */
		Symbol sym = Function.builtinFun(id);
		Symtab.getGlobalScope().insertSymLocal(sym);
	}

	public static boolean isBuiltin(Symbol sym) {
		return sym.getName().startsWith("__builtin_");
	}

	public static String stripBuiltinPrefix(String id) {
		if (!id.startsWith("__builtin_"))
			return id;
		
		for (String s : builtins)
			if (s.equals(id))
				return id;
		
		String s = id.substring(10);
		return s;
	}
	
	private static Symbol symForExpr(Expression e) {
		return e instanceof VarExpr ? ((VarExpr)e).getSymbol() : null;
	}

	public static boolean isBuiltin(Expression e) {
		Symbol sym = symForExpr(e);
		return (sym != null) ? isBuiltin(sym) : false;
	}

	public static boolean isVaArg(Expression e) {
		return symForExpr(e) == BI_VA_ARG;
	}

	public static Type typeOf(Expression e, Expression[] pars) {
		Symbol sym = symForExpr(e);
		
		if (sym == BI_VA_ARG) {
			/* The type of this is stored in a  sizeof in arg2 */
			return ((SizeofExpr)pars[1]).getSizeofType();
		}
		else if (sym == BI_OFFSETOF || 
				 sym == BI_VA_START ||
				 sym == BI_VA_END)
			return Type.Void;
		
		/* default return type for all other builtins */
		return Type.SignedInt;
	}

	public static StringBuffer PrintVaArg(Expression fun, Expression args[]) {
		Symbol sym = symForExpr(fun);
		assert (sym == BI_VA_ARG);
		
		StringBuffer sb = new StringBuffer();
		sb.append(BI_VA_ARG_NAME);
		sb.append("(");
		sb.append(ASTPrinter.toString(args[0], false));
		sb.append(",");
		sb.append(DeclPrinter.toString(((SizeofExpr)args[1]).type, ""));
		sb.append(")");
		return sb;
	}
}
