package crystal.emitc;

import crystal.ast.*;
import crystal.cfg.*;
import crystal.general.*;

import java.util.*;
import java.io.*;

/** Print all declaration in a symbol table. */
public class SymtabPrinter implements TypeVisitor<Object,Boolean> {

	static SymtabPrinter v = new SymtabPrinter();
	
	static Map<Symbol,Set<Symbol>> deps;
	static HashSymtab  scope;
	static Set<Symbol> tset, fset;
	
	static void initialize(HashSymtab s) {
		scope = s;
		deps = new HashMap<Symbol,Set<Symbol>>(); 
	}
	
	static public void getDeps(Symbol sym, boolean putdep) {
		tset = new HashSet<Symbol>();
		fset = new HashSet<Symbol>();
		
		Type type = sym.getType();
		if (type != null) {
			if (sym.getKind() == Flags.TYPE)
				type.accept(v, Boolean.FALSE);
			else
				type.accept(v, Boolean.TRUE);
		}
		
		if (putdep) {
			tset.addAll(fset);
			deps.put(sym, tset);
		}
	}
	
	public Object visit(PrimitiveType t, Boolean o) {
		return null;
	}
	
	public Object visit(ArrayType t, Boolean o) {
		t.getElementType().accept(this, o);
		
		for (Type et : typesInExpr(t.getExpression()))
			et.accept(this, Boolean.TRUE);

		return null;
	}
	
	/* Record struct symbols that occur inside params. */
	private static Stack<FunType> funStack = new Stack<FunType>();
	private static Set<Symbol> structsInParams = new HashSet<Symbol>();
	
	public Object visit(FunType t, Boolean o) {
		t.getReturnType().accept(this, o);
		
		funStack.push(t);
		for (int i = 0; i < t.numArgs(); ++i)
			t.getArg(i).accept(this, o);
		funStack.pop();
		return null;
	}
	
	public Object visit(PointerType t, Boolean o) {
		t.getTargetType().accept(this, Boolean.FALSE);
		return null;
	}
	
	public Object visit(AggregateType t, Boolean o) {
		for (int f = 0; f < t.numFields(); ++f) {
			Symbol field = t.getField(f);
			
			if (t.isEnum()) {
				Expression e = ((EnumConstant)field).getValue();
				if (e == null) continue;
				for (Type ty : typesInExpr(e)) 
					ty.accept(this, o);
			}
			else {
				field.getType().accept(this, o);
			}
		}
		return null;
	}
	
	public Object visit(AggrNameType t, Boolean o) {
		Set<Symbol> set = o == Boolean.TRUE ? tset : fset;
		Symbol sym = t.getTypeSymbol();
		
		if (set.contains(sym))
			return null;
		
		if (scope.lookupSymbol(sym) != sym)
			return null;

		if (!funStack.isEmpty())
			structsInParams.add(sym);
		
		if (o == Boolean.TRUE) {
			set.add(sym);

			Type st = t.getStructType();
			if (st != null) 
				st.accept(this, Boolean.TRUE);
		}
		
		return null;
	}

	public Object visit(TypedefType t, Boolean o) {
		Set<Symbol> set = o == Boolean.TRUE ? tset : fset;
		Symbol sym = t.getTypeSymbol();

		if (set.contains(sym))
			return null;

		if (scope.lookupSymbol(sym) != sym)
			return null;
		
		set.add(sym);

		if (o == Boolean.TRUE)
			t.getBaseType().accept(this, Boolean.TRUE);
		
		return null;
	}

	
	List<Type> typesInExpr(Expression e) {
		return typeHelper(e, new ArrayList<Type>());
	}

	private List<Type> typeHelper(Expression e, List<Type> vec) {
		if (e == null)
			return vec;
		
		if (e instanceof SizeofExpr) {
			SizeofExpr es = (SizeofExpr)e;
			vec.add(es.getSizeofType());
		}
		if (e instanceof VarExpr) {
			Symbol var = ((VarExpr)e).getSymbol();
			if (var instanceof EnumConstant) 
				vec.add(var.getType());
		}
		
		for (int k = 0; k < e.numExprKids(); k++)
			typeHelper(e.getExprKid(k), vec);

		return vec;
	}
	
	static Set<Symbol> visited;
	static Set<Symbol> traversed;
	static List<Symbol> list;
	
	private static List<Symbol> topSort() {
		list = new ArrayList<Symbol>();
		visited = new HashSet<Symbol>();
		traversed = new HashSet<Symbol>();

		for (Symbol sym : deps.keySet()) {
			assert(scope.lookupSymbol(sym) == sym);

			if (!visited.contains(sym))
				DFS(sym);
		}
		
		return list;
	}
	
	private static void DFS(Symbol sym) {
		visited.add(sym);
	
		for (Symbol kid : deps.get(sym)) 
			if (!visited.contains(kid))
				DFS(kid);

		list.add(sym);
	}
	
	public static void printGlobal(Symtab filetab, 
			PrintStream out, String ind) 
	{
		printSymtab(Symtab.getGlobalScope(), filetab, out, ind);
	}
	
	public static void printSymtab(HashSymtab symtab, 
			PrintStream out, String ind) 
	{
		printSymtab(symtab, null, out, ind);
	}
	
	private static void printSymtab(HashSymtab symtab, Symtab shadow,
			PrintStream out, String ind) 
	{
		if (symtab == null || symtab.isEmpty())
			return; /* Nothing to print */
	
		Set<Symbol> needed = null;
		if (shadow != null) 
			needed = getGlobalSymbols((FileSymtab)shadow);
		
		/* Print enums first. */
		for (Symbol sym: symtab.getSUESymbols()) {
			if (!(sym.getKind() == Flags.ENUM))
				continue;
			
			if (needed != null && !needed.contains(sym)) 
				continue;
			
			if (ind != null) out.print(ind);
			out.print(DeclPrinter.toString(sym.getType(), 
					sym.getName(), ind));
			out.println(";");
		}
		
		/* Collect typedef/struct dependences. */
		initialize(symtab);

		for (Symbol sym: symtab.getVFTSymbols()) 
			getDeps(sym, sym.getKind() == Flags.TYPE);
	
		for (Symbol sym: symtab.getSUESymbols())
			getDeps(sym, true);
				
		/* Declare structs that appear in function params. */
		for (Symbol sym : structsInParams) {
			out.print(Flags.kindToString(sym.getKind()) + " ");
			out.println(sym.getName() + ";");
		}
		structsInParams.clear();

		/* Print typedefs and structs in topological order. */
		for (Symbol sym : topSort()) {
			if (needed != null && !needed.contains(sym))
				continue;
				
			if (shadow != null && shadow.lookupSymbol(sym) != null)
				continue;

			if (sym.getKind() == Flags.ENUM)
				continue;
			
			if (ind != null) 
				out.print(ind);

			if (sym.getKind() == Flags.TYPE)
				out.print("typedef ");
			
			Type type = sym.getType();
			if (type == null) continue;
			
			out.print(DeclPrinter.toString(sym.getType(), 
					sym.getName(), ind));
			out.println(";");
		}
			
		if (symtab.getParent() == null) out.println();

		StringBuffer vars = null;
		StringBuffer funs = null;
		
		/* Print variables and functions. */
		for (Symbol sym: symtab.getVFTSymbols()) {
			if (needed != null && !needed.contains(sym)) 
				continue;
				
			if (shadow != null && shadow.lookupSymbol(sym) != null)
				continue;
			
			if (sym.getKind() == Flags.TYPE)
				continue;
			
			if (Builtins.isBuiltin(sym))
				continue;
			
			if (sym instanceof EnumConstant) 
				continue;
			
			StringBuffer sb = sym.isFunction() ? funs : vars;
			if (sb == null) {
				sb = new StringBuffer();
				if (sym.isFunction()) funs = sb; else vars = sb;
			}
			if (ind != null) sb.append(ind);
				
			if (!sym.isFunction() && sym.isUndefined())
				sb.append("extern ");
				
			if ((sym.getStorage() & Flags.STATIC) != 0)
				sb.append("static ");
				
			if ((sym.getStorage() & Flags.INLINE) != 0)
				sb.append("inline ");
				
			String name = sym.getIdName();
			sb.append(DeclPrinter.toString(sym.getType(), name, ind));
			sb.append(";\n");
		}
		
		if (funs != null) out.println(funs);
		if (vars != null) out.print(vars);
	}
	
	private static Set<Symbol> getGlobalSymbols(FileSymtab filetab) {
		Set<Symbol> syms = new HashSet<Symbol>();
		collect(filetab, syms);
		for (Function fun : Symtab.getFunctions(filetab.getFileName())) {
			collect(fun.getType(), syms);

			CFGInfo cfg = fun.getCFG();
			if (cfg == null) {
				collect(fun.getBody(), syms);
			}
			else {
				for (CFGNode n : cfg.getAllNodes()) 
					collect(n, syms);
			}
		}
		return syms;
	}

	private static void collect(CFGNode n, Set<Symbol> syms) {
		if (n instanceof CFGStatNode) {
			collect(((CFGStatNode)n).getStatement(), syms);
		}
		else {
			CFGSwitchNode s = (CFGSwitchNode) n;
			collect(s.getTestSrc(), syms);
			for (int i = 0; i < s.numEdges(); i++)
				collect(s.getLabelSrc(i), syms);
		}
	}

	private static void collect(Symtab tab, Set<Symbol> syms) {
		for (Symbol s : tab.getAllSymbols())
			collect(s.getType(), syms);
	
		for (int i = 0; i < tab.numKids(); i++)
			collect(tab.getKid(i), syms);
	}

	private static void collect(Type t, Set<Symbol> syms) {
		if (t == null) return;
		
		Symbol s = t.getTypeSymbol();
		if (s != null) {
			if (!syms.contains(s)) {
				syms.add(s);
				collect(s.getType(), syms);
			}
			else return;
		}
		
		if (t instanceof ArrayType) {
			ArrayType at = (ArrayType)t;
			collect(at.getExpression(), syms);
		}

		if (t.isEnum()) {
			AggregateType at = (AggregateType)t;
			for (int i = 0; i < at.numFields(); i++) {
				EnumConstant ec = (EnumConstant)at.getField(i);
				collect(ec.getValue(), syms);
			}
		}

		for (int i = 0; i < t.numTypeKids(); i++)
			collect(t.getTypeKid(i), syms);
	}

	private static void collect(Statement s, Set<Symbol> syms) {
    	if (s == null) return;

    	for (int i = 0; i < s.numExprKids(); i++)
			collect(s.getExprKid(i), syms);

		for (int i = 0; i < s.numStatKids(); i++)
			collect(s.getStatKid(i), syms);
	}

	private static void collect(Expression e, Set<Symbol> syms) {
    	if (e == null) return;

		if (e instanceof VarExpr) {
			Symbol s = ((VarExpr)e).getSymbol();
			syms.add(s);
			collect(s.getType(), syms);
		}

		for (int i = 0; i < e.numTypeKids(); i++)
			collect(e.getTypeKid(i), syms);

		for (int i = 0; i < e.numExprKids(); i++)
			collect(e.getExprKid(i), syms);

		for (int i = 0; i < e.numStatKids(); i++)
			collect(e.getStatKid(i), syms);
		
	}
}
