package crystal.general;

import java.io.*;
import java.util.*;

import crystal.ast.*;
import crystal.cfg.*;
import crystal.emitc.*;

public class Function extends Symbol {
	private static final long serialVersionUID = 3977863990124819766L;

	/** The formal function parameters. */
	private Symbol[] params;

	/** The return variable */
	private Symbol retvar;

	/** The shadow parameters */
	private Symbol[] shadows;

	/** The function body */
	private BlockStat body;

	/** The CFG constructed for this function */
	private CFGInfo cfg;

	/** The function symbol table */
	FunSymtab symtab;

	public Function(String name, FunType type, byte status, int storage, Loc loc) 
	{
		super(name, convertArrays(type), Flags.FUN, status, storage, loc);
		generateReturnVariable();
	}

	private static Type convertArrays(FunType type) {
		for (int i = 0; i < type.numArgs(); i++) {
			Type t = type.getArg(i).skipTypedefs();
			if (t.isArray()) {
				t = new PointerType(t.getTargetType());
				type.setArg(i, t);
			}
		}
		return type;
	}

	public static Function initFun(String name, String file) {
		Loc loc = Loc.newLoc(file, -1);
    	return new Function(name, Type.VoidToVoid, Flags.DEFINED, 0, loc);
	}
	
	public static Function builtinFun(String name) {
    	return new Function(name, Type.VoidToVoid, Flags.UNDEFINED, 0, Loc.UNK);
	}
	
	public static Function defaultFun(String name) {
		return new Function(name, Type.VoidToInt, Flags.NONE, 0, Loc.UNK);
	}

	private void generateReturnVariable() {
		Type rettype = ((FunType) type).getReturnType();
		if (rettype != Type.Void) 
			retvar = Symbol.newTemp(Options.SHADOW_RET_PREFIX + getName(), 
					                rettype);
		else
			retvar = null;
	}

	public String getIdName() {
		if (isAlloc()) return super.getName();
		
		return super.getIdName();
	}
	
	public boolean isFunction() {
		return true;
	}
	
	public boolean isInline() {
		return (storage & Flags.INLINE) != 0;
	}

	public void setBody(BlockStat body) {
		this.body = body;
	}

	public void setParams(List<Symbol> v) {
		this.params = new Symbol[v.size()];
		this.params = v.toArray(this.params);

		/* set param flags, fix types, generate the shadow variables */
		shadows = new Symbol[params.length];
		for (int p = 0; p < numParams(); ++p) {
			Symbol formal = getParam(p);
			formal.setParameter();
			
			/* convert array types into pointer types */
			Type t = formal.getType().skipTypedefs();
			if (t.isArray())
				formal.setType(new PointerType(t.getTargetType()));
			
			String sname = Options.SHADOW_ARG_PREFIX + formal.getName();
			Symbol shadow = Symbol.newTemp(sname, formal.getType());
			shadow.setShadow();
			shadows[p] = shadow;
		}
	}

	public void setInfoFrom(Function other) {
		this.type = other.type;
		this.params = other.params;
		this.shadows = other.shadows;
		this.body = other.body;
		this.cfg = other.cfg;
		this.symtab = other.symtab;
	}

	public BlockStat getBody() {
		return body;
	}

	public int numParams() {
		if (params == null)
			return 0;
		return params.length;
	}

	public Symbol getParam(int i) {
		return params[i];
	}

	public Symbol getReturnVar() {
		assert isDefined();
		return retvar;
	}

	public Symbol getShadowParam(int i) {
		return shadows[i];
	}

	public Iterable<Symbol> getSymbols() {
		return getSymTab().getVFTSymbols();
	}

	public CFGInfo getCFG() {
		return cfg;
	}

	public void setCFG(CFGInfo cfg) {
		this.cfg = cfg;
	}

	public List<CFGNode> getCFGNodes() {
		return cfg.getAllNodes();
	}
	
	public void printFunction(PrintStream ps, boolean useCFG) {
		if ((this.storage & Flags.STATIC) != 0)
			ps.print("static ");
		
		if ((this.storage & Flags.INLINE) != 0)
			ps.print("inline ");

		FunType ftype = (FunType) type;
		Type rettype = ftype.getReturnType();
		ps.print(DeclPrinter.toString(rettype, getIdName(), ""));
		ps.print("(");
		for (int i = 0; i < params.length; i++) {
			if (i > 0)
				ps.print(", ");
			Symbol arg = params[i];
			if (arg == Symbol.varargs)
				ps.print("...");
			else 
				ps.print(DeclPrinter.toString(arg.type, arg.getIdName(), ""));
		}
		ps.print(") ");
		if (cfg != null) 
			CFGToC.printC(ps, getSymTab(), cfg);
		else if (body != null)
			ps.print(ASTPrinter.toString(body, ""));
		else
			System.out.println(";");
		ps.println();
	}

	public HashSymtab getSymTab() {
		return (HashSymtab) symtab.getFirstKid();
	}

	public boolean isInit() {
		return getName().startsWith(Options.FILE_STARTUP_PREFIX);
	}

	public boolean isAlloc() {
		for (String a : Options.allocNames)
			if (getName().equals(a))
				return true;

		return false;
	}

	public boolean isFree() {
		for (String a : Options.freeNames)
			if (getName().equals(a))
				return true;

		return false;
	}
}