package crystal.projects.leakcontradictor;

import java.util.*;

import crystal.*;
import crystal.analysis.callgraph.CallGraph;
import crystal.analysis.ipa.*;
import crystal.analysis.pointer.*;
import crystal.ast.*;
import crystal.cfg.*;
import crystal.core.*;
import crystal.general.*;
import crystal.util.*;

/** The analysis class. */
public class Analysis implements IPAnalysis<E> {

	static CoreExpr nil = CoreExpr.zero;

	static Map<CFGNode,E> cache  = new HashMap<CFGNode,E>();
	Set<CFGNode> visited;

	/** Maximum number of steps per probe. 
	 *  Stop the trace afterwards.  */
	private static int MAX_STEPS = 500;

	/** Maximum number of regions in the region set. 
	 *  Stop the trace afterwards.  */
	private static int MAX_REGIONS = 50;
	
	/** Maximum number of calls per probe.
	 *  Skip function calls afterwards */
	private static int MAX_CALLS = 50;

	/** Maximum function call depth.
	 *  Skip function calls afterwards */
	private static int MAX_DEPTH = 1;

	/** If true, print the number of started and completed probes. */
	private static boolean showStats = false;

	
	/* Internal flags */
	private static boolean summarizeExtRefs = true;
	private static boolean doCache = false;
	private static boolean debug = false;

	/* We set this flag to limit the heuristic cut off to only happen 
	 * when we are actually transferring, and not doing simulated 
	 * transferring to setup the probe. */
	static boolean inProbe;

	
	static class TimeoutException extends RuntimeException {
		private static final long serialVersionUID = -2811798040505745053L;
	}

	/* Tests whether the expr has arrays/pointer arithmetic inside it*/
	static boolean hasArrays(CoreExpr expr) {
		if (expr == null)
			return false;

		if (expr instanceof CoreBinaryExpr)
			return true;

		for (int i = 0; i < expr.numKids(); i++)
			if (hasArrays(expr.getKid(i)))
				return true;

		return false;
 	}


	/* Tests whether cand occurs inside expr, as field or deref base */
	static boolean subAddr(CoreExpr expr, CoreExpr cand) {
		if (expr == null)
			return false;

		if (expr instanceof CoreDerefExpr) {
			CoreDerefExpr cde = (CoreDerefExpr)expr;
			if (cde.getExpr() == cand)
				return true;
		}

		if (expr instanceof CoreFieldExpr) {
			CoreFieldExpr cfe = (CoreFieldExpr)expr;
			if (cfe.getExpr() == cand)
				return true;
		}

		for (int i = 0; i < expr.numKids(); i++)
			if (subAddr(expr.getKid(i), cand))
				return true;

		return false;
	}

	/* Return whether this configuration is plausible if we assume expr 
	 * contains pointers that do not go anywhere */
	static boolean filterNullify(E e, CoreExpr expr) {
		if (expr == null)
			return true;

		for (CoreExpr he : e.secondary.hit)
			if (subAddr(he, expr))
				return false;

		/* ### FIXME: add miss expressions, if appropriate? */

		return true;
	}

	String lastFn = null;

	CoreExpr useSingleton(Region r) {
		CoreExpr rep = r.getInternalExpression();
		return CoreExpr.getVar(((CoreAddrExpr)rep).getSymbol());
	}

	/* Returns whether the expression is retainable given write to wr,
	 * with e1 decision e1d, and the original sign, sign; note that that
	 * for that to be -, it has to be explicit
	 * using rules [FILTER-FILTER3] */
	boolean retainable(CoreExpr expr, int sign, Region writeRegion, 
												CoreExpr e1, int e1d) {
		if (E.vstable(expr, writeRegion)) {
			return true; /* [FILTER1] */
		}

		if (E.vstable(e1, writeRegion) && 
            E.vstable(E.location(expr), writeRegion)) {

			if (e1d == -1 && sign == +1) {
				/* We wrote something that missed, hit after -> hit before 
				 * [FILTER2] */
				return true;
			}

			if (e1d == 1 && sign == -1) {
				/* We wrote something that hit, missed after -> miss before 
				 * [FILTER3] */
				return true;
			}
		}

		return false;
	}

	void transferAssign(CFGNode node, E e, Function fun,
			CoreDerefExpr e0, CoreExpr e1, Collection<E> out) {
		Region writeRegion = LCPt.regionOf(E.location(e0));
		int ld             = e.decision(e0);
		int rd             = e.decision(e1);

		/* Do not  try to trace things involving arrays -- heuristic */
		if (ld >= 0 && hasArrays(e1) || rd >= 1 && hasArrays(e0))
			return; //### FIXME: mark as abandoned via this.

		if (e.secondary.regions.size() > MAX_REGIONS && inProbe) {
			throw new TimeoutException();
		}
		
		boolean e0lstable = E.lstable(e0, writeRegion);

		if (e0lstable && E.vstable(e1, writeRegion)) {
			e = (E)e.clone(); //x5

			/* See whether we can remove any of one of the expressions 
			 * by assuming e0 == e1. hmm, this is actually symmetric! */
			Iterator<CoreExpr> i = e.secondary.regions.values().iterator();
			while (i.hasNext()) {
				CoreExpr rep = i.next();
				
				if (rep == nil)
					continue;
					
				CoreExpr s1 = rep.substitute(e0, e1);
				CoreExpr s2 = rep.substitute(e1, e0);
				
				if ((s1 != null && e.decision(s1) == -1) ||
					(s2 != null && e.decision(s2) == -1) ||
					e.decision(rep) == -1) {
						i.remove();
				}
			}
		}
 
		HashSet<CoreExpr> newHit  = new HashSet<CoreExpr>();
		HashSet<CoreExpr> newMiss = new HashSet<CoreExpr>();

		/*  Apply filter rules */
		for (CoreExpr he : e.secondary.hit)
			if (retainable(he, +1, writeRegion, e1, rd))
				newHit.add(he);

		for (CoreExpr me : e.secondary.miss)
			if (retainable(me, -1, writeRegion, e1, rd))
				newMiss.add(me);


		/* [SUBST1] */
		if (e0lstable) {
			for (CoreExpr he : e.secondary.hit) {
				if (E.vstable(he, writeRegion, e0)) {
					CoreExpr s = he.substitute(e0, e1);
					if (s != null)
						newHit.add(s);
				}
			}

			for (CoreExpr me : e.secondary.miss) {
				if (E.vstable(me, writeRegion, e0)) {
					CoreExpr s = me.substitute(e0, e1);
					if (s != null)
						newMiss.add(s);
				}
			}
			
			/* This is a somewhat trivial instantiation... 
			 * also includes [SUBST2]*/
			if (ld == +1)
				newHit.add(e1);
			if (ld == -1)
				newMiss.add(e1);
		}
		
		//### my rule..
		if (E.lstable(e1, writeRegion)) {
			if (rd == +1)
				newHit.add(e1);
			if (rd == -1)
				newMiss.add(e1);
		}

		/* Build a new configuration */
		E newC     = new E(e, newHit, newMiss);

		for (CoreExpr ce: newC.secondary.hit)
			assert( ce != null );

		for (CoreExpr ce: newC.secondary.miss)
			assert( ce != null );

		/* See whether any of the existing reference can be removed */
		if (e0lstable) {
			Iterator<Map.Entry<Region, CoreExpr>> i = 
				newC.secondary.regions.entrySet().iterator();
			while (i.hasNext()) {
				Map.Entry<Region, CoreExpr> pair = i.next();
				Region r = pair.getKey();
				if (r != writeRegion) {
					CoreExpr repE = pair.getValue();
					if (repE != nil) {
						boolean repMisses = (newC.decision(repE) == -1);
						//### rule
						if (!repMisses && 
							e1.isNumConstant() && subAddr(repE, E.location(e0)))
							repMisses = true;

						if (repMisses)
							i.remove();
					}
				}
			}
		}


		if (LCPt.regionOf(e0).equals(e.index.region)) {
			/* See whether we should add a new reference.. */
			CoreExpr prev = newC.secondary.regions.get(writeRegion);
			if (prev == null) /* First reference */
				newC.secondary.regions.put(writeRegion, e0);
			else if (prev != nil) { /* Had one reference*/
				if (e0lstable && (rd == +1 || newHit.contains(e1))) {
					newC.secondary.regions.put(writeRegion, e0);
				} else if (e0lstable && prev == e0) {
					/* We're writing to the one rep, can keep */
				} else { /*Give up!*/
					newC.secondary.regions.put(writeRegion, nil);
				}
			} else
				newC.secondary.regions.put(writeRegion, nil);
		} 

		/* Fix up variables to be singleton */
			
		for (Map.Entry<Region, CoreExpr> info : 
				 newC.secondary.regions.entrySet()) 
		{
			Region r = info.getKey();
			CoreExpr rep = info.getValue(); 
			if (rep == nil && LCPt.isSingletonRegion(r))
				info.setValue(useSingleton(r));
		}
		
		ArrayList<CoreExpr> toAdd = null;
		for (CoreExpr he : newC.secondary.hit) {
			CoreExpr loc = E.location(he);
			if (loc == null || loc.isNumConstant()) {
				continue;
			}
			Region   heR = LCPt.regionOf(loc);
			CoreExpr rep = newC.secondary.regions.get(heR);
			if (rep != nil && rep != null) {
				/* The rep of the region for hit expression must hit!*/
				if (toAdd == null)
					toAdd = new ArrayList<CoreExpr>();
				toAdd.add(rep);
			}
		}
		
		if (toAdd != null)
			for (CoreExpr a : toAdd)
				newC.secondary.hit.add(a);

		/* Cleanup: remove all implicit miss expressions */
		Iterator<CoreExpr> i = newC.secondary.miss.iterator();
		while (i.hasNext()) {
			CoreExpr mi = i.next();
			if (newC.isImplicitMinus(mi))
				i.remove();
		}

		/* Check whether it's feasible */
		//### Should do that w/o making a new config?
		newC.fillInIndex();
		if (!newC.applyConstraints())
			return;
		
		out.add(newC);
	}

	static ArrayList<CoreDerefExpr> allPointerLocals(Function function) {
		ArrayList<CoreDerefExpr> allPtLocals = nonFormalPointerLocals(function);
		allPtLocals.addAll(formalPointerLocals(function));
		return allPtLocals;
	}
	
	static boolean isVar(Symbol s) {
		return (s.getKind() == Flags.VAR) &&
			  !(s instanceof EnumConstant);
	}
	
	static void addGlobals(Symtab symtab, ArrayList<CoreDerefExpr> out) {
		if (symtab instanceof HashSymtab) {
			HashSymtab tab = (HashSymtab)symtab;
			Iterable<Symbol> vars = tab.getVFTSymbols();
			for (Symbol s : vars) {
				if (isVar(s) && s.getStorage() == Flags.STATIC) {
					CoreDerefExpr expr = CoreExpr.getVar(s);
					if (!LCPt.noPointers(expr))
						out.add(expr);
				}
			}
		}
		
		for (Symtab kid : symtab.getKids())
			addGlobals(kid, out);
	}
	
	static ArrayList<CoreDerefExpr> allPointerGlobals(Function function) {
		ArrayList<CoreDerefExpr> allPtGlobals = new ArrayList<CoreDerefExpr>();
		addGlobals(Symtab.getGlobalScope(), allPtGlobals);
		return allPtGlobals;
	}

	static ArrayList<CoreDerefExpr> formalPointerLocals(Function function) {
		ArrayList<CoreDerefExpr> allPtLocals = new ArrayList<CoreDerefExpr>();

		for (int pn = 0; pn < function.numParams(); ++pn) {
			CoreDerefExpr expr = CoreExpr.getVar(function.getParam(pn));
			if (!LCPt.noPointers(expr))
				allPtLocals.add(expr);
		}

		return allPtLocals;
	}

	static ArrayList<CoreDerefExpr> nonFormalPointerLocals(Function function) {
		ArrayList<CoreDerefExpr> allPtLocals = new ArrayList<CoreDerefExpr>();

		Iterable<Symbol> local =
				function.getSymTab().getVFTSymbols();
		for (Symbol s : local) {
			if (isAutomatic(s)) {
				CoreDerefExpr expr = CoreExpr.getVar(s);
				if (!LCPt.noPointers(expr))
					allPtLocals.add(expr);
			}
		}
		return allPtLocals;
	}

	public Collection<E> transfer(CFGNode node, E e, Function function) {
		if (doCache) {
			E cachedItem = cache.get(node);
			if (cachedItem != null && cachedItem.isGeneralizationOf(e)) {
				System.err.println("**** cache hit ****");
				return new ArrayList<E>();
			}
			
			visited.add(node);
		}

		if (debug && !function.toString().equals(lastFn)) {
			lastFn = function.toString();
			System.err.println("fun:" + lastFn);
		}
		
		if (debug)
			System.err.println("node:" + node + "\n\t in :" + e);
			
		Collection<E> out = new ArrayList<E>();

		if (node instanceof CFGAssignNode) {
			CFGAssignNode a = (CFGAssignNode)node;
			transferAssign(node, e, function, a.getLeft(), a.getRight(), out);
			return out;
		}

		if (node instanceof CFGCallNode) {
			CFGCallNode cn = (CFGCallNode)node;
			if (cn.isAlloc()) {
				CoreDerefExpr lhs = cn.getLeft();

				/* Filter out based on field's being null */
				Collection<CoreExpr> exprs = getFieldExprs(lhs);

				if (exprs != null)
					for (CoreExpr fe : exprs)
						if (!filterNullify(e, fe))
							return out;

				if (lhs != null) {
					if (hasArrays(lhs))
						return out; /* Heurstically ignore array-related stuff*/

					/* Compute whether we have many refs under the assumption of
					 * lhs + */
					boolean manyRefs = false;
					Region lhsRegion = LCPt.regionOf(E.location(lhs));
						
					for (CoreExpr he : e.secondary.hit)
						if (!LCPt.regionOf(E.location(he)).equals(lhsRegion))
							manyRefs = true;

					if (e.secondary.guaranteedExternRef)
						manyRefs = true;
						
					/* If yes, better add a miss expression, this can't hit! */
					if (manyRefs) {
						e = (E)e.clone();
						e.secondary.miss.add(lhs);
					}

					int ad = e.decision(lhs);

					if (ad == +1 && manyRefs)
						return out; /* Bottom */
					else if (ad == -1) {
						transferAssign(node, e, function, lhs, nil, out);
					} else {
						Traces.warning(
							(Traces.total == 0 ? "\n" : "") +
							"Warning: potential leak:\n" +
							"Error location: " + 
							Traces.probeOriginNames() +
							"Allocated at: \t" + Traces.nodeInfo(cn),
							e);
					}
				} else {
					/* malloc value not used -- we ignore this here, 
					 * warn on init */
					out.add(e);
				}

				return out;
			}
		}

		if (node instanceof CFGReturnNode) {
			/* For return node, we handle this as a nullification
			 * of &_returnMarker preceeding _returnMarker = return argument */
			CFGReturnNode ret = (CFGReturnNode)node;
			if (ret.getExpr() != null) {
				/* Grab the return slot */
				CoreDerefExpr slot = getReturnVar(function);

				transferAssign(node, e, function,
					slot, ret.getExpr(), out);

				Iterator<E> i = out.iterator();
				while (i.hasNext()) {
					E cand = i.next();
					if (!filterNullify(cand, slot.getExpr()))
						i.remove();
					else
						cand.secondary.regions.remove(
									LCPt.regionOf(slot.getExpr()));
				}

				return out;
			}
		}

		if (node.isExit()) {
			/* At exit, all the locals get nullified, simulate that  */
			ArrayList<CoreDerefExpr> toNullify = allPointerLocals(function);

			/* Compute effects */
			out.add(e);
			for (CoreDerefExpr toNull : toNullify) {
				ArrayList<E> out2 = new ArrayList<E>();
				for (E in : out)
					transferAssign(node, in, function,
						toNull, nil, out2);

				out = out2;
			}

			/* Remove the entry with no refcounts we put in as a starting point */
			///### FIXME: this is broken, with a hack to make it suck less
			Iterator<E> iter = out.iterator();
			while (iter.hasNext()) {
				E entry = iter.next();
				if (entry.secondary.depth == 0 && 
					entry.secondary.regions.size() == 0) {
					iter.remove();
				}
			}

			return out;
		}

		/* By default, just pass though */
		out.add(e);
		return out;
	}

	public boolean filter(E e, CoreExpr test, boolean b) {
		if (test instanceof CoreUnaryExpr) {
			CoreUnaryExpr ule = (CoreUnaryExpr)test;
			if (ule.getOp() == LogicalOps.NOT)
				return filter(e, ule.getExpr(), !b);
		}

		assert !e.secondary.hit.contains(nil);

		CoreExpr testLeft  = null;
		CoreExpr testRight = null;
		int      testOp    = -1;
		
		/* Turn if (foo) into if (foo != 0) */
		if (test instanceof CoreDerefExpr) {
			testLeft  = test;
			testRight = nil;
			testOp    = CompareOps.NEQ;
		}
		
		/* Handle cases like if (0) or do { } while(0); --- impossible edges
		 * ### generalize */
		if (test instanceof CoreConstExpr) {
			if (test.isInteger(0) && b)
				return false;
		}

		/* Extract out normal tests */
		if (test instanceof CoreBinaryExpr) {
			CoreBinaryExpr cbe = (CoreBinaryExpr)test;
			testLeft  = cbe.getLeft();
			testRight = cbe.getRight();	
			testOp    = cbe.getOp();
		}

		/* Handle &&, || if any left. 
		 * Can also handle not cases, with an or I think */
		if (testOp == LogicalOps.AND && b == true) {
			/* Both must be true on true branch */
			return  filter(e, testLeft, true) &&
					filter(e, testRight, true);
		}

		if (testOp == LogicalOps.OR && b == false) {
			/* Both must be false on false branch */
			return filter(e, testLeft, false) &&
			       filter(e, testRight, false);
		}
		
		if ((testOp == CompareOps.EQ  && b) ||
			(testOp == CompareOps.NEQ && !b)) {
				return filterEqual(e, testLeft, testRight);
		}
		
		if ((testOp == CompareOps.EQ  && !b) ||
			(testOp == CompareOps.NEQ && b)) {

			/* We got that 2 things are not equal */
			if (e.decision(testRight) == 1)
				e.secondary.miss.add(testLeft);

			if (e.decision(testLeft) == 1)
				e.secondary.miss.add(testRight);
		}

		if (e.applyConstraints()) {
			e.fillInIndex();
			return true;
		} else {
			return false;
		}
	}

	public boolean filterEqual(E e, CoreExpr testLeft, CoreExpr testRight) {
		int ld = e.decision(testLeft);
		if (ld == 1)
			e.secondary.hit.add(testRight);
		else if (ld == -1)
			e.secondary.miss.add(testRight);

		int rd = e.decision(testRight);
		if (rd == 1)
			e.secondary.hit.add(testLeft);
		else if (rd == -1)
			e.secondary.miss.add(testLeft);

		/* Try substitutions on hit/miss sets */
		ArrayList<CoreExpr> toAdd = new ArrayList<CoreExpr>();
		for (CoreExpr he : e.secondary.hit) {
			CoreExpr s1 = he.substitute(testLeft, testRight);
			CoreExpr s2 = he.substitute(testRight, testLeft);
			if (s1 != null && s1 != he && s1 != testRight)
				toAdd.add(s1);

			if (s2 != null && s2 != he && s2 != testLeft)
				toAdd.add(s2);
		}

		e.secondary.hit.addAll(toAdd);

		toAdd.clear();
		for (CoreExpr me : e.secondary.miss) {
			CoreExpr s1 = me.substitute(testLeft, testRight);
			CoreExpr s2 = me.substitute(testRight, testLeft);
			if (s1 != null && s1 != me && s1 != testRight)
				toAdd.add(s1);

			if (s2 != null && s2 != me && s2 != testLeft)
				toAdd.add(s2);
		}

		e.secondary.miss.addAll(toAdd);
		if (e.applyConstraints()) {
			e.fillInIndex();
			return true;
		} else {
			return false;
		}
	}

	static CoreDerefExpr getReturnVar(Function fun) {
		Symbol sym = fun.getReturnVar();
		if (sym != null)
			return CoreExpr.getVar(sym);
		else
			return null;
	}

	static CoreDerefExpr getArgSlot(Function fun, int arg) {
		CoreDerefExpr cde = CoreExpr.getVar(fun.getShadowParam(arg));
		return cde;
	}

	boolean isSafeToRetain(CoreExpr expr, Function callee) {
		if (expr instanceof CoreDerefExpr) {
			CoreDerefExpr fe = (CoreDerefExpr)expr;
			CoreExpr base = fe.getExpr();
			if (base instanceof CoreAddrExpr) {
				CoreAddrExpr ae = (CoreAddrExpr)base;
				if (!ae.getSymbol().hasAddressTaken())
					return true;
			}
		}

		return false;
	}


	/* Filter out locals not visible in the callee. */
	HashSet<CoreExpr> filterLocal(E c, Set<CoreExpr> in, boolean isHit, 
								  Set<CoreDerefExpr> ignore, Function callee) 
    {
		HashSet<CoreExpr> out = new HashSet<CoreExpr>();
		for (CoreExpr expr: in) {
			if (!usesLocals(expr, ignore, callee) && 
				!isSafeToRetain(expr, callee)) {
				out.add(expr);
			}
		}

		return out;
	}

	/* Add back filtered-out locals */
	void restoreFiltered(Set<CoreExpr> in, Set<CoreExpr> out,
						 Set<CoreDerefExpr> ignore, Function callee) {
		for (CoreExpr expr: in) {
			if (isSafeToRetain(expr, callee))
				out.add(expr);
		}
	}

	Pair<Iterable<E>, Iterable<E>> bypass(Iterable<E> out) {
		return new Pair<Iterable<E>,Iterable<E>>(null, out);
	}

	Pair<Iterable<E>, Iterable<E>> mapped(Iterable<E> out) {
		return new Pair<Iterable<E>,Iterable<E>>(out, null);
	}

	
	public Pair<Iterable<E>, Iterable<E>> 
	map(CFGCallNode callsite, E postCall, Function callee) 
	{
		ArrayList<E> candidates = new ArrayList<E>();

		/* See whether we're dealing with free. */ 
		if (callsite.isFree()) {
			CoreExpr deletee = callsite.getArg(0);
			/* Do we hit the configuration? */
			int dd = postCall.decision(deletee);
			if (dd == -1)
				candidates.add(postCall);

			if (dd == 0) {
				E newE = (E)postCall.clone();
				newE.secondary.miss.add(deletee);
				candidates.add(newE);
			}
			
			/* +1 is contradiction, so nothing added */
			return bypass(candidates);
		}
		
		/* See whether the function is special... */
		CoreExpr dest = callsite.getFun();
		
		if (dest instanceof CoreDerefExpr)
			dest = ((CoreDerefExpr)dest).getExpr();
		
		if (dest instanceof CoreAddrExpr) {
			CoreAddrExpr cae = (CoreAddrExpr)dest;
			
			boolean realloc = cae.getSymbol().getName().equals("realloc");
			boolean strcpy  = cae.getSymbol().getName().equals("strcpy");
			
			if (realloc || strcpy) {
				CoreDerefExpr left = callsite.getLeft();
				if (left == null) {
					if (realloc)
						Traces.warning("Warning: Unused value of realloc at " 
							+ Traces.nodeInfo(callsite), null);
					/* in case of strcpy with left = right, nothing to 
						worry about */
				} else {
					/* Model as foo = bar to get the new reference in there */
					transferAssign(callsite, postCall, callee, left,
									callsite.getArg(0), candidates);
				}
				return bypass(candidates);
			}
		}

		/* Bypass the call if need be.. */
		if (callee.getCFG() == null || 
				postCall.secondary.depth == MAX_DEPTH || 
				postCall.secondary.calls == MAX_CALLS) {
			candidates.add(postCall.clone());
			return bypass(candidates);
		}

		postCall = (E)postCall.clone();

		/* Do substitution/assignment unless we have no left-hand-side on call,
		 * or configuration is a special one saying that the return misses, 
		 * in which case we setup things differently. */
		boolean doSubst = callsite.getLeft() != null && 
						  !postCall.index.missedByReturn;
		CoreDerefExpr rSlot = getReturnVar(callee);

		if (doSubst && rSlot == null)
			/* We are potential calling a function that returns void, 
			 * and using the return value! No need to warn here, though, 
			 * the probe setup code will. */
			doSubst = false;

		/* Add proper miss for return slot */
		if (postCall.index.missedByReturn && rSlot != null)
			postCall.secondary.miss.add(rSlot);

		if (doSubst) {
			/* Return sequence is rougly: 
			 * lhs = retMarker; retMarker = 0 */
			transferAssign(callsite, postCall, callee,
				rSlot, nil, candidates);

			ArrayList<E> out2 = new ArrayList<E>();

			for (E e : candidates)
				transferAssign(callsite, e, callee,
					callsite.getLeft(), rSlot, out2);

			candidates = out2;
		} else {
			/* .. or just put in the unmodified thingie */
			candidates.add(postCall);
		}

		/* Fix depth, maps of candidates */
		for (E e: candidates) {
			e.secondary.depth++;
			e.secondary.calls++;
		}

		Set<CoreDerefExpr> exclude = new HashSet<CoreDerefExpr>();
		exclude.add(rSlot);
		for (E e : candidates) {
			for (CoreExpr he : e.secondary.hit) {
				if (isSafeToRetain(he, callee))
					e.secondary.guaranteedExternRef = true;
			}

			S sec = e.secondary;
			sec.hit  = filterLocal(e, sec.hit,  true,  exclude, callee);
			sec.miss = filterLocal(e, sec.miss, false, exclude, callee);

			/* Remove/summarize extern regs */
			if (summarizeExtRefs) {
				ArrayList<Region> toRem = new ArrayList<Region>();
				for (Region r : e.secondary.regions.keySet())
					if (LCPt.isLocalRegion(r)) {
						e.secondary.guaranteedExternRef = true;
						//System.err.println("kill:" + r);
						toRem.add(r);
					}

				for (Region r: toRem)
					e.secondary.regions.remove(r);
			}

			/* Hide those that would cause trouble */
			for (Region r : e.secondary.regions.keySet()) {
				CoreExpr old = e.secondary.regions.get(r);
				if (old != nil && usesLocals(old, exclude, callee))
					e.secondary.regions.put(r, nil);
			}
		}
		
		/* See whether we have any arguments that will  be bound to the
		 * actuals all the way to the end, and if so, we can apply an equality 
		 * filter to all of them */
		for (int p = 0; p < callee.numParams() && p < callsite.numArgs(); ++p) {
			Symbol formal = callee.getParam(p);
			if (LCPt.isStableFormal(formal)) {
				CoreExpr arg = callsite.getArg(p);
				CoreExpr formalCore = CoreExpr.getVar(formal);
				
				/* See whether the actual is: 
				 * (1) unchanged 
				 * (2) free of symbols that would cause problems in the callee */
				if (isSafeToRetain(arg, callee) && 
				    !usesLocals(arg, exclude, callee)) {
					Iterator<E> i = candidates.iterator();
					while (i.hasNext()) {
						E cand = i.next();
						if (!filterEqual(cand, arg, formalCore)) {
							i.remove();
						}
					}
				}
			}
		}

		return mapped(candidates);
	}


	HashSet<CoreExpr> replace(HashSet<CoreExpr> set, CoreExpr oldE, CoreExpr newE) {
		HashSet<CoreExpr> outSet = new HashSet<CoreExpr>();

		for (CoreExpr ce : set) {
			CoreExpr s = ce.substitute(oldE, newE);
			if (s != null)
				outSet.add(s);
			else
				outSet.add(ce);
		}
		
		return outSet;
	}

	ArrayList<E> transferAll(CFGNode node, ArrayList<E> in,
				Function fun, CoreDerefExpr lhs, CoreExpr rhs) {
		ArrayList<E> out = new ArrayList<E>();
		for (E e : in)
			transferAssign(node, e, fun, lhs, rhs, out);
		return out;
	}

	/* Out-of-scopes all the variables, making sure to get rid of the */
	ArrayList<E> filterNullifyAll(ArrayList<E> in, CoreExpr filter) {
		CoreDerefExpr fde = (CoreDerefExpr)filter;
		ArrayList<E> out = new ArrayList<E>();
		for (E e : in)
			if (filterNullify(e, fde.getExpr()))
				out.add(e);
		return out;
	}

	public CoreDerefExpr getParam(Function fun, int pos) {
		CoreDerefExpr cde = CoreExpr.getVar(fun.getParam(pos));
		return cde;
	}
	
	/* The unmapping function.  */
	public Collection<E> unmap(CFGCallNode call, E postCall, E atEntry,
								Function callee) {
		ArrayList<E> confs = new ArrayList<E>();

		/* Kill locals which aren't formals */
		ArrayList<CoreDerefExpr> toNull = nonFormalPointerLocals(callee);
		for (CoreDerefExpr expr : toNull) {
			CoreExpr base = expr.getExpr();
			if (!filterNullify(atEntry, base)) {
				return confs;
			}
		}

		confs.add(atEntry);

		/* Get all the shadows, and zero-fill them -- they're entering 
		 * the scope.  */
		for (int p = 0; p < callee.numParams(); ++p) {
			confs = transferAll(call, confs, callee,
								getArgSlot(callee, p), nil);
		}

		/* formal = shadow */
		for (int p = 0; p < callee.numParams(); ++p)
			confs = transferAll(call, confs, callee,
				getParam(callee, p), getArgSlot(callee, p));

		/* now, formals go out of scope */
		for (int p = 0; p < callee.numParams(); ++p) 
			confs = filterNullifyAll(confs, getParam(callee, p));

		/* Add back removed stuff */
		if (postCall != null)
			for (E e: confs) {
				restoreFiltered(postCall.secondary.hit,  e.secondary.hit,  null, callee);
				restoreFiltered(postCall.secondary.miss, e.secondary.miss, null, callee);

				/* Add extern regs */
				if (summarizeExtRefs) 
					for (Region r : postCall.secondary.regions.keySet()) 
						if (LCPt.isLocalRegion(r)) {
							e.secondary.regions.put(r, useSingleton(r));
						}
				e.secondary.guaranteedExternRef = 
									postCall.secondary.guaranteedExternRef;
			}

		int limit = Math.min(callee.numParams(), call.numArgs());

		/* shadow = actual */
		for (int p = 0; p < limit; ++p)
			confs = transferAll(call, confs, callee,
				getArgSlot(callee, p), call.getArg(p));

		/* out-of-scope the shadows */
		for (int p = 0; p < limit; ++p) 
			confs = filterNullifyAll(confs, getArgSlot(callee, p));

		/* fix up depth */
		for (E c : confs) {
			if (c.secondary.depth != 0)
				--c.secondary.depth;
		}
		
		return confs;
	}

	private static Map<CoreExpr, ArrayList<CoreExpr>> fieldExprs = 
		new HashMap<CoreExpr, ArrayList<CoreExpr>>();
	
	private static void collectFieldExpressions() {
		for (CoreExpr e : CoreExpr.allExprs.keySet()) {
			if (e instanceof CoreDerefExpr) {
				CoreDerefExpr e1 = (CoreDerefExpr)e;
				CoreExpr      e2 = e1.getExpr();
				
				if (e2 instanceof CoreFieldExpr) {
					CoreFieldExpr cfe = (CoreFieldExpr)e2;
					Symbol        fld = cfe.getField();
					
					if (fld != null) {
						ArrayList<CoreExpr> fl = fieldExprs.get(cfe);
						if (fl == null) {
							fl = new ArrayList<CoreExpr>();
							fieldExprs.put(cfe.getExpr(), fl);
						}
						fl.add(cfe);
					}
				}
			}
		}
	}

	@SuppressWarnings("unchecked")
	static ArrayList<CoreExpr> getFieldExprs(CoreDerefExpr expr) {
		ArrayList<CoreExpr> fmiss = fieldExprs.get(expr);

		if (fmiss != null)
			fmiss = (ArrayList<CoreExpr>)fmiss.clone();

		/* Also add any fields based on the type of the expression*/
		ArrayList<CoreExpr> derefExprs = getDerefFieldExprs(expr);

		if (derefExprs != null) {
			if (fmiss == null)
				fmiss = new ArrayList<CoreExpr>();

			for (CoreExpr ce : derefExprs) {
				CoreDerefExpr cde = (CoreDerefExpr)ce;
				fmiss.add(cde.getExpr());
			}
		}
		return fmiss;
	}

	static ArrayList<CoreExpr> getDerefFieldExprs(CoreDerefExpr expr) {
		ArrayList<CoreExpr> fmiss = null;

		/* Add any fields based on the type of the expression*/
		Type t = expr.getType();
		if (t instanceof PointerType) {
			Type it = ((PointerType)t).getTargetType().unwrap();

			if (it instanceof AggregateType) {
				AggregateType st = (AggregateType)it;
				if (fmiss == null)
					fmiss = new ArrayList<CoreExpr>();

				for (int fn = 0; fn < st.numFields(); ++fn) {
					Symbol   f    = st.getField(fn);
					CoreExpr fLoc = CoreExpr.getFieldExpr(expr, f);
					fmiss.add(CoreExpr.getDerefExpr(fLoc, f.getType()));
				}
			} /* struct */
		} /* ptr */
		return fmiss;
	}

	static int probes      = 0;
	static int timeouts    = 0;
	
	static Function testFun;

	static IPAnalysisEngine<I,S,E,A> eng;

	static void testConfig(Analysis analysis, Function fun, CFGNode node, E e, 
			boolean atExit) { 
		if (debug) {
			System.err.println("--- probe start ---  :" + e + 
					"in function " + fun + "/" + atExit + " node:" + node);
		}

		++probes;

		eng = new IPAnalysisEngine<I,S,E,A>(analysis,
											IPAnalysisEngine.BACKWARD);
											
		eng.setKeepLinks();
		

		I.flushIndexTable();
		e.fillInIndex();

		Traces.probeStarted(eng, node, e);

		analysis.visited = new HashSet<CFGNode>();

		boolean timeout = false;

		if (!atExit)
			eng.initialize(fun, node, e);
		else
			eng.initializeAtSpot(fun, node, e);
		try {
			inProbe = true;
			eng.setCutoff(MAX_STEPS);
			int iterations = eng.run();
			/* Note timeout from the IPA engine limiting..*/
			if (iterations >= MAX_STEPS) 
				timeout = true;
			inProbe = false;
		}
		catch (TimeoutException err) {
			timeout = true;
			inProbe = false;
		}

		if (timeout)
			++timeouts;
	}

	static void initConfig(Analysis analysis, CoreDerefExpr lhs,
						   CFGNode node, Function fun, CoreExpr rhs) {
		Region writeRegion = LCPt.regionOf(E.location(lhs));
		Region confRegion  = LCPt.regionOf(lhs);

		/* Create the element, and feed it */
		Traces.probeOrigin(node);
		E e = new E(confRegion, writeRegion, lhs);
		if (rhs != null)
			e.secondary.miss.add(rhs);

		testConfig(analysis, fun, node, e, false);
	}

	static boolean isAutomatic(Symbol s) {
		return isVar(s) && (s.getStorage() != Flags.STATIC);
	}

	static void initFieldClear(Analysis an, CoreExpr ce, CFGNode node, 
								Function fun, Collection<CoreExpr> exprs) {
		CoreDerefExpr cde = (CoreDerefExpr)ce;
		if (!LCPt.noPointers(cde)) {
			Region writeRegion = LCPt.regionOf(E.location(cde));
			Region confRegion  = LCPt.regionOf(cde);
	
			/* Create the element... */
			Traces.probeOrigin(node);
			E e = new E(confRegion, writeRegion, cde);

			/* Add additional regions for other fields which may 
			 * point to this. */
			for (CoreExpr fld : exprs) {
				if (LCPt.regionOf(fld).equals(confRegion)) {
					Region reg = LCPt.regionOf(E.location(fld));
					if (e.secondary.regions.get(reg) != fld)
						e.secondary.regions.put(reg, nil);
					else
						e.secondary.regions.put(reg, fld);
				}
			}

			testConfig(an, fun, node, e, false);
		}
	}
	
	protected static String escape(String data) {
		StringBuffer result = new StringBuffer();
		for (int i = 0; i < data.length(); ++i) {
			char c = data.charAt(i);
			switch(c) {
				case '_': result.append("\\_"); break;
				default:   result.append(c);
			}
		}
		return result.toString();
	}
	
	static int cfgCount = 0;

	static void probeFunctionInfo(final Analysis analysis, final Function fun) {
		CFGWalker init = new CFGWalker() {
			public void visit(CFGStatNode node) {
				Traces.newProbe();
				if (node instanceof CFGAssignNode) {
					CFGAssignNode assign = (CFGAssignNode)node;

					CoreDerefExpr lhs = assign.getLeft();
					if (LCPt.noPointers(lhs) || hasArrays(lhs) || 
							hasArrays(assign.getRight())) {
						return; /* Nothing to do here */
					}

					initConfig(analysis, lhs, node, fun, assign.getRight());
				}

				if (node instanceof CFGCallNode) {
					CFGCallNode call = (CFGCallNode)node;

					if (call.getLeft() != null) {
						/* See whether we know the callees. If we don't, 
						 * we assume this is stuff like malloc, rand, and
						 * just pass things through. Otherwise, we
						 * set things up for unmap. Unfortunately,
						 * we can't do both, so we pray it doesn't happen
						 * ### the proper thing might be for transfer to
						 * fudge things the right way.
						 * 
						 * (Of course, if we now the destination is an 
						 * allocator, we treat it as such) */
						boolean callsUnknown = false;
						boolean noTrueReturn = false;
						for (Function callee : CallGraph.getCallees(call)) {
							if (callee.getCFG() == null)
								callsUnknown = true;
							else if (getReturnVar(callee) == null) {
								Traces.warning("Potential call to function " +
									 callee + 
									 " returning void using return value" + 
									 " at statement " + call + 
									 " in function" + fun, null);
								noTrueReturn = true;
							}
						}

						/* Handle the bypass/init case. This is a bit weird 
						 * in the case of assignment not actually happening, 
						 * but we warn about that, do.. */
						if (callsUnknown || call.isAlloc() || noTrueReturn) {
							CoreDerefExpr lhs = call.getLeft();
							if (LCPt.noPointers(lhs) || hasArrays(lhs))
								return;
							
							initConfig(analysis, lhs, node, fun, null);
						}
						
						/* Handle case of return value clobering.. */
						if (!(callsUnknown || call.isAlloc())) {
							/* Feed configuration -after- the call, setting 
							 * things up as if we were in the middle of unmap.*/
							CoreDerefExpr lhs = call.getLeft();
							if (LCPt.noPointers(lhs)) return;
							
							Region writeRegion = LCPt.regionOf(E.location(lhs));
							Region confRegion  = LCPt.regionOf(lhs);
							E e = new E(confRegion, writeRegion, lhs, true);
							Traces.probeOrigin(node);
						
							/* Place it right after this node.. */
							testConfig(analysis, fun, node, e, true);
						}
					} else {
						/*### CHECK THIS, due to IR changes */
						if (call.isAlloc())
							Traces.warning("Warning: alloc value not "
							+ "used in " + node + " in function" + fun, null);
						if (call.isFree()) {
							/* Free leaks everything referred to by fields 
							 * of the expression */
							CoreExpr del = call.getArg(0);
							if (!(del instanceof CoreDerefExpr)) {
								if (Options.VERBOSITY >= 2)
									System.out.println("Unable to model:" + call);
								return;
							}
							
							Collection<CoreExpr> exprs =
								getDerefFieldExprs((CoreDerefExpr)del);

							if (exprs != null)
								for (CoreExpr ce : exprs)
									initFieldClear(analysis, ce,
										node, fun, exprs);
						}
					}
				}

				if (node.isExit()) {
					/* At exit, all the locals get nullified. We simulate 
					 * this from various program points */
					ArrayList<CoreDerefExpr> locs = allPointerLocals(fun);
					
					if (fun.getName().equals("main")) {
						/* At exit of main all globals get nullified */
						ArrayList<CoreDerefExpr> locs2 = allPointerGlobals(fun);
						for (CoreDerefExpr cde : locs2)
							locs.add(cde);
					}

					/* First, iterate over all possible starting points */
					outer: for (CoreDerefExpr leakSource : locs) {
						/* Create configuration for this */
						Traces.newProbe();
						Traces.probeOrigin("End of scope of \"" + 
								leakSource + "\"" + Traces.locationInfo(node));
						E e = new E(LCPt.regionOf(leakSource),
									LCPt.regionOf(E.location(leakSource)),
									leakSource);
						e.fillInIndex();

						/* Then compute the effect of transfers that happen
						 * 'before' it */
						boolean sawIt = false;
						ArrayList<E> out = new ArrayList<E>();
						for (CoreDerefExpr other : locs) {
							if (other == leakSource)
								sawIt = true;
							else if (sawIt) {
									analysis.transferAssign(node, e, fun, 
											other, nil, out);
								if (out.isEmpty())
									continue outer;
								
								e = out.get(0);
								for (int oi = 1; oi < out.size(); ++oi)
									e.secondary.merge(out.get(oi).secondary);
								out.clear();
							}
						}

						/* Looks plausible thus far -- try it out */
						testConfig(analysis, fun, node, e, false);
					}
					
					/* We also feed a configuration hit by return marker
					 * if there are functions that do not use our return value
					 * ### FIXME: return foo when return type is void should be
					 * handled as plain argument-less return
					 */
					CoreDerefExpr slot = getReturnVar(fun);
					if (slot == null || LCPt.noPointers(slot))
						return;
					
					ArrayList<CFGCallNode> calls = new ArrayList<CFGCallNode>();

					Iterable<CFGCallNode> cs = CallGraph.getCallSitesTo(fun);
					if (cs == null) return;

					for(CFGCallNode callNode: cs)
						if (callNode.getLeft() == null)
							calls.add(callNode);

					if (!calls.isEmpty()) {
						E e = new E(LCPt.regionOf(slot),
									LCPt.regionOf(E.location(slot)),
									slot);

						for (int i = 1; i < calls.size(); ++i)
							Traces.probeOrigin(calls.get(i));

						testConfig(analysis, fun, node, e, true);
					}
				}
			}
		};
		init.visitAll(fun.getCFG());
	}
	
	/* Run the analysis */
	public static void run() {		
		collectFieldExpressions();
		LCPt.init();
		Traces.init();

		/* Build the analysis engine object. */
		final Analysis analysis = new Analysis();

		/*  initialize the dataflow facts appropriately, and test stuff */
		for (final Function fun : Symtab.getFunctions()) {
			if (fun.isInit()) continue;
			testFun = fun;
			probeFunctionInfo(analysis, fun);
		}
	}

	/* Helper to determine whether the given core expression uses 
	 * non-shadow locals */
	static class ExprScopeDeterminer extends RecursiveCoreExprVisitor {
		boolean             sawLocal;
		Set<CoreDerefExpr>  nonLocal;
		Function            callee;
		static  ExprScopeDeterminer self = new ExprScopeDeterminer();

		static boolean usesLocals(CoreExpr expr, Set<CoreDerefExpr> nonLocal, 
				Function callee) 
		{
			self.callee   = callee;
			self.sawLocal = false;
			self.nonLocal = nonLocal;
			expr.accept(self, null);
			return self.sawLocal;
		}

		public Object visit(CoreDerefExpr expr, Object obj) {
			if (nonLocal != null && nonLocal.contains(expr))
				return null;
			else
				return super.visit(expr, obj);
		}

		public Object visit(CoreAddrExpr expr, Object obj) {
			Symbol sym = expr.getSymbol();
		
			/* It may be tempting to filter out things like local statics 
			 * here as well, but that's not right: they may be useful in 
			 * case of recursion, including indirect one */
			//FIXME: BUG: BUG: BUG: messes up file scope
			if (isAutomatic(sym) && 
					callee.getSymTab().lookupVFTRecurse(sym.getName()) == sym &&
					Symtab.getGlobalScope().lookupVFTLocal(sym.getName()) != sym) {
				sawLocal = true;
			}

			return null;
		}
	}
	
	static boolean usesLocals(CoreExpr e, Set<CoreDerefExpr> nonLocal, 
			Function callee) {
		return  ExprScopeDeterminer.usesLocals(e, nonLocal, callee);
	}

	static public void main(String[] args) {
		Options.CFG = true;
		Options.CALL_GRAPH = true;
		Options.ALLOC_WRAPPER_DETECTION = true;
		Options.POINTER_ANALYSIS = true;
		PointerAnalysis.STEENSGAARD = true;
		Main.main(filterArgs(args));

		System.out.println();
		out("Leak contradictor analysis");

		if (Options.VERBOSITY >= 2) {
			out("Parameters:");
			out("  Max steps   = "  + MAX_STEPS);
			out("  Max regions = "  + MAX_REGIONS);
			out("  Max calls   = "  + MAX_CALLS);
			out("  Max depth   = "  + MAX_DEPTH);
		}
		
		out("Starting analysis...");
		long time = System.currentTimeMillis();
		Analysis.run();
		time = System.currentTimeMillis() - time;

		if (showStats) {
			out(String.format("Probes: %d total", probes));
			out(String.format("Probes: %d completed (%.0f%%)",
					(probes-timeouts), 
					100.0*(float)(probes-timeouts)/(float)probes));
		}
		
		if (Traces.total == 0 && Options.VERBOSITY <= 2)
			out("No memory leaks found.");
		else {
			if (Traces.hasBugDB) {
				out("Bugs: " + Traces.validWarn);
				out("False positives: " + Traces.invalidWarn);
				out("Unclassified: " + Traces.unknownWarn);
			}
			else {
				out("Total: "  + Traces.total + 
					" potential memory leak" + (Traces.total > 1 ? "s" : "") +
					" found.");
			}
			
			out("Traces printed to file \"" + Traces.traceFile + "\"");
		}
		
		out(String.format("Analysis time: %.2f s", (float)(time/1000.0)));
	}
	
	private static void out(String s) {
		System.out.println("[LC] " + s);
	}


	/** Parse the comand line arguments. */
	static String[] filterArgs(String[] args) {
		List<String> newargs = new ArrayList<String>();
		
		for(int i = 0; i < args.length; i++) {
			String arg = args[i];
			if (arg.startsWith("-trace-file=")) {
				Traces.traceFile = value(args[i]);
			}
			else if (arg.startsWith("-v=")) {
				try {
					Options.VERBOSITY = Integer.parseInt(value(args[i]));
				} catch (Exception e) {
					error("option -v requires a number.\n");
				}
			} 
			else if (arg.startsWith("-max-steps=")) {
				try { 
					MAX_STEPS = Integer.parseInt(value(args[i]));
				} catch (Exception e) {
					error("option -max-steps requires a number.\n");
				}
			} 
			else if (arg.startsWith("-max-regions=")) {
				try { 
					MAX_REGIONS = Integer.parseInt(value(args[i]));
				} catch (Exception e) {
					error("option -max-regions requires a number.\n");
				}
			} 
			else if (arg.startsWith("-max-calls=")) {
				try { 
					MAX_CALLS = Integer.parseInt(value(args[i]));
				} catch (Exception e) {
					error("option -max-calls requires a number.\n");
				}
			} 
			else if (arg.startsWith("-max-depth=")) {
				try { 
					MAX_DEPTH = Integer.parseInt(value(args[i]));
				} catch (Exception e) {
					error("option -max-depth requires a number.\n");
				}
			} 
			else if (arg.equals("-show-stats")) {
				showStats = true;
			} 
			else if (arg.startsWith("-alloc=")) {
				for (String name : value(args[i]).split(","))
					Options.allocNames.add(name);
			} 
			else if (arg.startsWith("-free=")) {
				for (String name : value(args[i]).split(","))
					Options.freeNames.add(name);
			} 
			else if (arg.startsWith("-")) {
				error(String.format("option %s not supported.", arg)); 
			} 
			else  {
				newargs.add(arg);
			}
		}

		if (newargs.size() == 0)
			usage();
		
		String[] res = new String[newargs.size()];
		return newargs.toArray(res);
	}
		
	private static String value(String s) {
		return s.substring(s.indexOf("=")+1);
	}


	/** Print error message and exit. */
	static void error(String msg) {
		System.out.println("lc : " + msg);
		System.exit(1);
	}
	
	/** Print the command line usage. */
	static void usage() {
		System.out.printf(
		"Usage: lc [options] [preprocessed C files]\n" +
		"Options (default values shown in parentheses):\n"  +
		"   -trace-file=<f> the leak report file (\"leak-traces.txt\")\n" +
		"   -max-steps=<n>  max steps per probe, then stop tracing (500)\n" +
		"   -max-regset=<n> max region set size, then stop tracing (50)\n" +
		"   -max-calls=<n>  max calls per probe, then skip calls (50)\n" +
		"   -max-depth=<n>  max fun. call depth, then skip calls (1)\n" +
		"   -show-stats     print the number of started/completed probes\n" +
		"   -alloc=<funcs>  specify additional allocator functions\n" +
		"   -free=<funcs>   specify additional free functions\n"
		);

		System.exit(0);
	}
}

/* kate: indent-width 4; replace-tabs off; tab-width 4; space-indent off;*/
