package crystal.projects.leakcontradictor;

import crystal.core.*;
import crystal.analysis.ipa.*;
import crystal.analysis.pointer.*;

import java.util.*;

/** The abstraction element. */
public class E extends Element<I,S,E> {
	boolean missedByReturn;
	Region  ourRegion;

	private E() {
	}

	/* Note that none of the constructors for E fill in the index. 
	 * This must be done separately when the initialization is done */
	public E(E other, HashSet<CoreExpr> newHit, HashSet<CoreExpr> newMiss) {
		missedByReturn = false;
		ourRegion      = other.ourRegion;

		context = other.context;
		secondary = new S(newHit, newMiss, 
                             other.secondary.regions);
		secondary.depth = other.secondary.depth;
		secondary.calls = other.secondary.calls;
		secondary.guaranteedExternRef = other.secondary.guaranteedExternRef;
	}

	public E(Region ourRegion, Region lastReference, CoreExpr hit) {
		this(ourRegion, lastReference, hit, false);
	}

	public E(Region ourRegion, Region lastReference, CoreExpr hit, 
			boolean missedByReturn) 
	{
		this.missedByReturn = missedByReturn;
		this.ourRegion      = ourRegion;

		secondary    = new S();
		secondary.regions.put (lastReference, hit);
		secondary.hit.add     (hit);
		context      = null;
	}
	
	public boolean isGeneralizationOf(E other) {
		/* If we're more general, we should have less hit/miss expressions */
		for (CoreExpr he : secondary.hit)
			if (!other.secondary.hit.contains(he))
				return false;
				
		for (CoreExpr me : secondary.miss)
			if (!other.secondary.miss.contains(me))
				return false;
				
		/* ... but more regions ... */
 		for (Region r : secondary.regions.keySet()) {
 			if (!other.secondary.regions.containsKey(r))
 				continue;
 				
 			/* Other has this region, too. We better not have 
 			   non-nil where other has nil or other
 			*/
 			CoreExpr rep = secondary.regions.get(r);
 			if (rep == Analysis.nil)
 				continue;
 				
 			CoreExpr oRep = other.secondary.regions.get(r);
 			if (oRep != rep)
 				return false;
 		}
 		
 		for (Region r : other.secondary.regions.keySet())
 			if (!secondary.regions.containsKey(r))
 				return false;
 		
		return true;
	}

	void fillInIndex() {
		HashSet<Region> hSet = new HashSet<Region>();
		for (CoreExpr he : secondary.hit) {
			CoreExpr loc = location(he);
			if (loc == null || loc.isNumConstant())
				continue;
				
			if (loc != null)
				hSet.add(LCPt.regionOf(loc));
		}
			
		HashSet<Region> mSet = new HashSet<Region>();
		for (CoreExpr me : secondary.miss) {
			CoreExpr loc = location(me);
			if (loc == null || loc.isNumConstant())
				continue;
				
			if (loc != null)
				mSet.add(LCPt.regionOf(loc));
		}
		
		index = I.newI(ourRegion, hSet, mSet, this.missedByReturn);
	}

	public E clone() {
		E e = new E();
		e.ourRegion = ourRegion;
		e.index     = index;
		e.context   = context;
		e.secondary = secondary.clone();
		return e;
	}

	int decision(CoreExpr e) {
		/* Unwrap casts --- if this was a pointer hitting something
		  it still would be, and vice-versa */
		while (e instanceof CoreUnaryExpr) {
			CoreUnaryExpr cue = (CoreUnaryExpr)e;
			if (cue.isCast())
				e = cue.getExpr();
			else break;
		}

		if (secondary.hit.contains(e))
			return +1;

		if (isMinus(e))
			return -1;
		
		return 0;
	}

	boolean isPlus(CoreExpr e) {
		if (secondary.hit.contains(e))
			return true;
		return false;
	}

	boolean isMinus(CoreExpr e) {
		if (secondary.miss.contains(e) ||
		    isImplicitMinus(e))
			return true;
		
		return false;
	}
	
	
	static boolean derefsConstant(CoreExpr expr) {
		if (expr == null)
			return false;

		if (expr instanceof CoreDerefExpr) {
			CoreDerefExpr cde = (CoreDerefExpr)expr;
			if (cde.getExpr().isNumConstant())
				return true;
		}

		if (expr instanceof CoreFieldExpr) {
			CoreFieldExpr cfe = (CoreFieldExpr)expr;
			if (cfe.getExpr().isNumConstant())
				return true;
		}

		for (int i = 0; i < expr.numKids(); i++)
			if (derefsConstant(expr.getKid(i)))
				return true;

		return false;
	}

	boolean isImplicitMinus(CoreExpr e) {
		//### FIXME: other cases - symbolic constants,
		//references of ints, use points-to information?
		return  e.isNumConstant() ||
			derefsConstant(e) ||
			zeroRC(e);
	}

	boolean zeroRC(CoreExpr e) {
		CoreExpr loc = location(e);
		if (loc == null) {
			return true; //Region.noPointers(e);
		}
		
		Region reg = LCPt.regionOf(loc);
		if (secondary.regions.containsKey(reg))
			return false;
		
		return true;
	}

	static CoreExpr location(CoreExpr e) {
		if (!(e instanceof CoreDerefExpr))
			return null;
		
		return ((CoreDerefExpr)e).getExpr();
	}
	
	static boolean vstable(CoreExpr e, Region r, CoreExpr ignore) {
		if (e == null)
			return true;

		if (e == ignore)
			return true;

		if (e instanceof CoreDerefExpr) {
			CoreDerefExpr cde  = (CoreDerefExpr)e;
			CoreExpr      expr = cde.getExpr();
			Region        reg  = LCPt.regionOf(expr);

			if (r.equals(reg))
				return false;
		}
		for (int i = 0; i < e.numKids(); i++)
			if (!vstable(e.getKid(i), r, ignore))
				return false;
		
		return true;
	}

	static boolean vstable(CoreExpr e, Region r) {
		return vstable(e, r, null);
	}
	
	static boolean lstable(CoreExpr e, Region r) {
		return vstable(location(e), r);
	}

	static boolean lstable(CoreExpr e, Region r, CoreExpr ignore) {
		return vstable(location(e), r, ignore);
	}

	/* Applies constraints, and returns true if the configuration survives */
	boolean applyConstraints() {
		/* Check whether we're both minus and plus! */
		for (CoreExpr h : secondary.hit) {
			if (isMinus(h))
				return false;
		}

		return true;
	}


	public String toStringDot() {
		return "{" + index.toStringDot() + "|" + secondary.toStringDot() + "}";
	}

	public String toString() {
		return toStringDot();
	}
}
