package crystal.projects.leakcontradictor;

import crystal.analysis.pointer.*;
import crystal.ast.*;
import crystal.cfg.*;
import crystal.core.*;
import crystal.general.*;

import java.util.*;

/* Wrappers around points-to info, or address-taken + type */
class LCPt {
	static ParamStable ps = new ParamStable();
	static boolean noPt = true;

	static Region everythingElse = null;

	static class TypeRegion extends Region
	{
			boolean isSingletonRegion;
			boolean isLocalRegion;
			Type    t;
			Symbol  f;

			TypeRegion(CoreExpr reg, Type t) {
				super(reg);
				this.t = t;
			}

			TypeRegion(CoreExpr reg, Symbol f) {
				super(reg);
				this.f = f;
			}


			public String toString() {
				if (t != null)
					return t.toString();
				else if (f != null)
					return f.toString();
				else {
					return getInternalExpression().toString();
				}
			}

			TypeRegion(CoreExpr reg, boolean local) {
				super(reg);
				isSingletonRegion = true;
				isLocalRegion     = local;
			}

			public boolean equals(Object o) {
				if (!(o instanceof TypeRegion))
						return false;

				TypeRegion other = (TypeRegion)o;
				return getInternalExpression() == other.getInternalExpression();
			}
	}

	static void init() {
		if (!noPt) return;
		AddressTaken.init();
		ps.init();
	}

	static Type normalizeType(Type type) {
		if (type instanceof ArrayType) {
			ArrayType at = (ArrayType)type;
			return new ArrayType(
				normalizeType(at.getElementType()), 
							  at.getExpression());
		} 

		if (type instanceof PrimitiveType) {
			if (((PrimitiveType)type).isInteger())
				return Type.SignedInt;
			return type;
		}

		if (type instanceof FunType) {
			return type; //### FIXME
		}

		if (type instanceof PointerType) {
			PointerType pt = (PointerType)type;
			return new PointerType(
				normalizeType(pt.getTargetType()), pt.getQuals());
		}

		if (type instanceof AggrNameType) {
			AggrNameType snt = (AggrNameType)type;
			return snt.mostSpecificType();
		}

		if (type instanceof AggregateType) {
			AggregateType st = (AggregateType)type;
			return st.mostSpecificType();
		}

		if (type instanceof TypedefType) {
			TypedefType tt = (TypedefType)type;
			return tt.mostSpecificType();
		}

		throw new RuntimeException("huh " + type);
	}

	static HashMap<String, Region> typeNodes  = new HashMap<String, Region>();
	static HashMap<Symbol, Region> fieldNodes = new HashMap<Symbol, Region>();

	static Region regionOf(CoreExpr e) {
		if (noPt) {
			/* See whether this is a var w/o addresstaken */
			if (e instanceof CoreAddrExpr) {
				CoreAddrExpr addr = (CoreAddrExpr)e;
				if (!addr.getSymbol().hasAddressTaken()) {
					/* Note that we have to distinguish between the notion 
					 * of singleton and non-address taken variable here. */
					return new TypeRegion(e, isLocal(addr.getSymbol()));
				} 
			}

			/* See whether this is a field w/o address */
			if (e instanceof CoreFieldExpr) {
				CoreFieldExpr cfe = (CoreFieldExpr)e;
				if (!AddressTaken.isFieldAddressTaken(cfe.getField())) {
					Symbol key = cfe.getField();
					Region   r   = fieldNodes.get(key);
					if (r == null) {
						Symbol sym = Symbol.newTemp(cfe.getField().getName(), 
								                    Type.Char);
						r = new TypeRegion(CoreExpr.getAddressOf(sym), sym);
						fieldNodes.put(key, r);
					}
					return r;
				}
			}

			/* Get everything else a node based on its type */
			Type   t = normalizeType(e.getType());
			Region r = typeNodes.get(t.toString());
			if (r == null) {
				Symbol sym  = Symbol.newTemp(t.toString(), Type.Char);
				r = new TypeRegion(CoreExpr.getAddressOf(sym), t);
				typeNodes.put(t.toString(), r);
			}

			return r;
		}
		else
			return Region.regionOf((CoreDerefExpr)e);
	}

	static boolean noPointers(CoreExpr e) {
		if (noPt) {
			return !(normalizeType(e.getType()) instanceof PointerType);
		} else
			return Region.noPointers((CoreDerefExpr)e);
	}

	/*### this should probably move someplace else --- duplicated from analysis */
	private static boolean isAutomatic(Symbol s) {
		return(s.getStorage() != Flags.STATIC && s.getKind() == Flags.VAR);
	}

	private static boolean isLocal(Symbol sym) {
		/*### CHECKME: File scope ? */
		if (isAutomatic(sym) && 
				Symtab.getGlobalScope().lookupVFTLocal(sym.getName()) != sym)
			return !sym.hasAddressTaken();
		return false;
	}

	static boolean isSingletonRegion(Region reg) {
		if (noPt)
			return ((TypeRegion)reg).isSingletonRegion;

		CoreExpr rep = reg.getInternalExpression();
		if (rep instanceof CoreAddrExpr) {
			Symbol sym = ((CoreAddrExpr)rep).getSymbol();
			return !sym.hasAddressTaken();
		}

		return false;
	}
	
	static boolean isLocalRegion(Region reg) {
		if (noPt)
			return ((TypeRegion)reg).isLocalRegion;

		CoreExpr rep = reg.getInternalExpression();
		if (rep instanceof CoreAddrExpr) {
			Symbol sym = ((CoreAddrExpr)rep).getSymbol();
			return isLocal(sym);
		}

		return false;
	}
	
	static boolean isStableFormal(Symbol sym) {
		return !ps.unstableFormals.contains(sym);
	}
}

class AddressTaken extends CFGWalker {
	static boolean initialized = false;

	static Set<Symbol> fieldTaken = new HashSet<Symbol>();

	public static boolean isFieldAddressTaken(Symbol sym) {
		assert initialized;
		return fieldTaken.contains(sym);
	}

	public static void init() {
		AddressTaken analysis = new AddressTaken();

		for (Function fun: Symtab.getFunctions())
			analysis.visitAll(fun.getCFG());
		
		initialized = true;
	}

	ExprChecker exprChecker  = new ExprChecker();

	public void visit(CFGStatNode node) {
		exprChecker.process(node.getStatement());
	}

	public void visit(CFGSwitchNode node) {
		exprChecker.process(node.getTestSrc());
	}
	
	class ExprChecker extends SimplePass {
		public void localVisitPre(AddressofExpr expr) {
			Expression base = expr.getBase();
			if (base instanceof FieldExpr) {
				FieldExpr fe = (FieldExpr)base;
				fieldTaken.add(fe.getField());
			}
		}
	}
}

class ParamStable extends CFGWalker
{
	HashSet<Symbol> unstableFormals = new HashSet<Symbol>();
	HashSet<Symbol> curFormals;
	
	
	void init() {
		long time = System.currentTimeMillis();
		 
		for (Function fun: Symtab.getFunctions()) {
			/* Fill in the list of formals..*/
			curFormals = new HashSet<Symbol>();
			for (int p = 0; p < fun.numParams(); ++p) {
				Symbol par = fun.getParam(p);
				curFormals.add(par);
				if (par.hasAddressTaken())
					unstableFormals.add(par);
			}
			
			visitAll(fun.getCFG());
		}
		
		time = System.currentTimeMillis() - time;
	}
	
	public void visit(CFGAssignNode node) {
		CoreExpr leftBase = node.getLeft().getExpr();
		if (leftBase instanceof CoreAddrExpr) {
			Symbol sym = ((CoreAddrExpr)leftBase).getSymbol();
			if (curFormals.contains(sym)) {
				unstableFormals.add(sym);
			}
		}
    }

}


