package crystal.projects.leakcontradictor;

import crystal.analysis.callgraph.*;
import crystal.analysis.ipa.*;
import crystal.ast.*;
import crystal.cfg.*;
import crystal.general.*;
import crystal.util.*;

import java.io.*;
import java.util.*;

class Traces {
	static String traceFile = "leak-traces.txt";
	static String bugdbFile = "Classify.lc";
	
	static PrintStream traceOutput;

	/* State needed for diagnostics... */
	static HashMap<CFGNode, Function> nodeFunctions =
					new HashMap<CFGNode, Function>();

	/** Gets information on location of the node for use in warnings */
	static String locationInfo(CFGNode node) {
		StringBuilder build = new StringBuilder();
		
		ASTNode infoNode = null;
		if (node instanceof CFGStatNode)
			infoNode = ((CFGStatNode)node).getStatement();
		else
			infoNode = ((CFGSwitchNode)node).getTestSrc();
			
		String fileName = infoNode.getFile();
		if (fileName == null)
			fileName = "(unknown)"; 
		/* Should get removed when the builder preserves all info */
			
		build.append("\n\t\tin file \"");
		build.append(new File(fileName).getName());
		build.append("\", function \"");
		build.append(nodeFunctions.get(node).toString());
		build.append("\", line ");
		
		int line = infoNode.getLeftLine();
		if (node.isExit())
			line = infoNode.getRightLine(); /* End of block..*/
		
		build.append(line);
		build.append("\n");
		return build.toString();
	}

	/** Gets information on the node for use in warnings */
	static String nodeInfo(CFGNode site) {
		StringBuilder build = new StringBuilder();
		build.append(site);
		build.append(locationInfo(site));
		return build.toString();
	}
		
	static boolean hasBugDB = true;

	/* This map contains information on how to mark the bug 
	   from the classifications file..*/
	static HashMap<String, Classification> classifyWarnings = 
		new HashMap <String, Classification>();
	
	static HashSet<String> issued = new HashSet<String>();

	enum Classification {
		Bug,
		FalsePositive
	};

	static int total = 0;
	static int unknownWarn = 0;
	static int validWarn   = 0;
	static int invalidWarn = 0; 
	
	/* Combined worklist and trace entry... */
	static class TraceLink {
		CFGNode   node;
		E         conf;
		TraceLink origin;

		TraceLink(CFGNode _node, E _conf) {
			node = _node;
			conf = _conf;
			origin = null;
		}

		TraceLink(CFGNode _node, E _conf, TraceLink _origin) {
			node = _node;
			conf = _conf;
			origin = _origin;
		}
	}

	/* Returns all predecessors for a node, including IPA ones */
	static ArrayList<CFGNode> fullPredecessors(TraceLink trace) {
		CFGNode node = trace.node;
		
		ArrayList<CFGNode> preds = new ArrayList<CFGNode>();
		for (int k = 0; k < node.numPreds(); k++)
			preds.add(node.getPred(k));

		if (node.isCall()) {
			CFGCallNode call = (CFGCallNode)node;
			Iterable<Function> callees = CallGraph.getCallees(call);
			if (callees != null)
				for (Function callee : callees)
					if (callee.getCFG() != null)
						preds.add(callee.getCFG().exitNode);
		}

		if (node.isEntry()) {
			Function f = nodeFunctions.get(node);
			
			/* See if we can find out what call we came from so 
			   we don't build a trace with an unfeasible path */
			CFGNode knownCall = null;
			Pair<TraceLink, TraceLink> callInfo = findMatchingCall(trace);
			if (callInfo != null)
				knownCall = callInfo.snd.node;
				
			if (f != null) {
				Iterable<CFGCallNode> callers = CallGraph.getCallSitesTo(f);
				if (callers != null) {
					for (CFGCallNode call : callers) {
						if (knownCall != null && call != knownCall)
							continue;
						
						TraceLink tmp = new TraceLink(call, null, trace);
						preds.addAll(fullPredecessors(tmp));
					}
				}
			}
		}

		/* Filter stuff that has no info..*/
		ArrayList<CFGNode> out = new ArrayList<CFGNode>();
		for (CFGNode cand : preds) {
			if (probeEng.getLinkInfo(cand).size() > 0)
				out.add(cand);
		}
		
		return out;
	}

	static HashSet<Pair<CFGNode, E>> traceVisited = 
		new HashSet<Pair<CFGNode, E>>();
			
	static HashMap<Pair<CFGNode, E>, HashSet<Pair<CFGNode, E>>> summaries =
		new HashMap<Pair<CFGNode, E>, HashSet<Pair<CFGNode, E>>>();
					
	static void addTraceSummary(TraceLink exitNode, TraceLink entryNode) {
		if (exitNode.conf == null) return; /* Was synthesized from summary 
											  already */
		Pair<CFGNode, E> exitKey = 
			new Pair<CFGNode, E>(exitNode.node, exitNode.conf);
		HashSet<Pair<CFGNode, E>> info = summaries.get(exitKey);
		
		if (info == null) {
			info = new HashSet<Pair<CFGNode, E>>();
			summaries.put(exitKey, info); 
		}
		
		info.add(new Pair<CFGNode, E>(entryNode.node, entryNode.conf));
	}
	
	static HashSet<Pair<CFGNode,E>> getTraceSummary(CFGNode exitNode,
													E exitConf) {
		Pair<CFGNode, E> exitKey = new Pair<CFGNode, E>(exitNode, exitConf);
		return summaries.get(exitKey);
	}

	static void addWorkListEntry(LinkedList<TraceLink> worklist,
								TraceLink link, boolean force)
	{
		Pair<CFGNode, E> dupeCheck = new Pair<CFGNode, E>(link.node, link.conf);
		if (!force && traceVisited.contains(dupeCheck)) return;
		traceVisited.add(dupeCheck);
		worklist.addLast(link);
	}
	
	static void addWorkListEntry(LinkedList<TraceLink> worklist,
								TraceLink link) {
		addWorkListEntry(worklist, link, false);
	}
	
	/* Returns the pair of exit node and call corresponding to this entry node,
	 * or null if there is none. */
	static Pair<TraceLink, TraceLink> findMatchingCall(TraceLink entryNode) {
		int level = 0;
		for (TraceLink cur = entryNode; cur != null; cur = cur.origin) {
			if (cur.node.isEntry())
				++level;
			/* Note that we match the exit node and note the call node 
			   since some calls may be to external functions and hence 
			   not paired with anything */
			if (cur.node.isExit() && cur.origin != null) {
				--level;
				if (level == 0)
					return new Pair<TraceLink, TraceLink>(cur, cur.origin);
			}
		}
		
		/* May happen if we started in the beginning of the call */
		return null;
	}
	
	static void printConstructedTrace(TraceLink cur, CFGNode startNode) {
		if (traceOutput == null) {
			System.out.println("Unable to store trace");
			return;
		}
	
		traceOutput.println("Trace:");
		traceOutput.printf(" %-20s  %-4s\n", "Function", "Line");
		
		/* First, the trace will often lack the last node 
		 * (depending on probe placement). Link it in if that's the case */
		TraceLink last = cur;
		while (last.origin != null)
			last = last.origin;
			
		if (last.node != startNode)
			last.origin = new TraceLink(startNode, null);
			
		/* Further, we visit the call nodes in funny order since we are 
		 * tracing backwards. Fix it if we can */
		TraceLink prev   =  null;
		TraceLink cursor =  cur;
		while (cursor != null) {
			if (cursor.node.isEntry()) {
				Pair<TraceLink, TraceLink> callInf = findMatchingCall(cursor);
				if (callInf != null) {
					/* Link in call in front of us, link exit through */
					TraceLink exit = callInf.fst;
					TraceLink call = callInf.snd;
					
					exit.origin = call.origin;
					call.origin = cursor;
					if (prev == null)
						cur = call;
					else
						prev.origin = call;
				}
			}
		
			prev   = cursor;
			cursor = cursor.origin; 
		}
		
		int indent = 1;
		while (cur != null) {
			if (indent < 1) indent = 1;
			
			/* Output function and line name */
			String funName = "";
			Function curFun = nodeFunctions.get(cur.node);
			if (curFun != null)
				funName = curFun.getName();
			
			int line = -1;
			ASTNode astNode;
			if (cur.node instanceof CFGStatNode)
				astNode = ((CFGStatNode)cur.node).getStatement();
			else
				astNode = ((CFGSwitchNode)cur.node).getTestSrc();
				
			if (astNode.getFile() != null)
				line = astNode.getLeftLine();
			
			traceOutput.printf(" %-20s  %-4s", funName, line == -1 ? "" : line);
			
			for (int t = 0; t < indent; ++ t)
				traceOutput.print("     ");

			if (cur.node.isEntry()) {
				Function fun = nodeFunctions.get(cur.node);
				
				traceOutput.print("[Called " + fun + "]");
				++indent;
			} else if (cur.node.isExit()) {
				traceOutput.print("[Returning]");
				--indent;
			} else if (cur.node instanceof CFGSwitchNode) {
					CFGSwitchNode sw = (CFGSwitchNode)cur.node;
					if (sw.isIfThenElse()) {
						traceOutput.print("if " + 
							ASTPrinter.toString(sw.getTestSrc(),true)); 
					} else {
						traceOutput.print(cur.node);
					}
			} else {
				traceOutput.print(cur.node);
			}
			
			if (cur.node instanceof CFGSwitchNode && cur.origin != null) {
				CFGSwitchNode sw = (CFGSwitchNode)cur.node;
				
				for (int e = 0; e < sw.numEdges(); ++e) {
					if (sw.getEdge(e) == cur.origin.node) {
						if (!sw.isIfThenElse())
							traceOutput.print(" [took label:" +
								sw.getLabelSrc(e) + "]");
						else if (sw.getLabelSrc(e) == null)
							traceOutput.print(" [took branch:true]");
						else
							traceOutput.print(" [took branch:false]");
					}
				}
			}

			traceOutput.println();
			
			cur = cur.origin;
		}
		traceOutput.println("\n");
	}
	
	static void outputTrace(E endConf) {
		LinkedList<TraceLink> worklist = new LinkedList<TraceLink>();
		traceVisited.clear();

		TraceLink start = new TraceLink(probeStartNode, probeStartElement);
		addWorkListEntry(worklist, start);

		/* We may also need to start off one level off due to the initial
		   propogate. We know the things are proper based off prev = null
		*/
		ArrayList<CFGNode> initalPreds = fullPredecessors(start);
		for (CFGNode pred : initalPreds) {
    		List<Pair<E,E>> links = probeEng.getLinkInfo(pred);
			for (Pair<E,E> link : links) {
				if (link.fst == null)
					addWorkListEntry(worklist, new TraceLink(pred, link.snd));
			}
		}

		 
		/* Do BFS over link structure to find end node */
		while (!worklist.isEmpty()) {
			//System.err.println("sz:" + worklist.size());
			TraceLink cur = worklist.removeFirst();
			if (cur.conf  == endConf) {
				printConstructedTrace(cur, probeStartNode);
				return;
			} else {
				if (cur.node.isEntry()) {
					/* Store summary information if possible */
					Pair<TraceLink, TraceLink> callInfo = findMatchingCall(cur);
					if (callInfo != null)
						addTraceSummary(callInfo.fst, cur);
				}
				
				/* See whether any of the predessors are relevant.
				   Predecessors here are not just intra-procedural, but
				   also inter-procedural */
				
				ArrayList<CFGNode> preds = fullPredecessors(cur);
				for (CFGNode pred : preds) {
					List<Pair<E,E>> predLinks = probeEng.getLinkInfo(pred);
					if (predLinks == null) continue;

					for (Pair<E,E> cand : predLinks) {
						if (cand.fst == cur.conf) {
							addWorkListEntry(worklist, new TraceLink(pred, cand.snd, cur));
							
							/* May also need to use summary cache in case 
							   we're re-exploring the call */
							if (pred.isExit()) {
								HashSet<Pair<CFGNode,E>> summary = 
												getTraceSummary(pred, cand.snd);
								if (summary != null) {
									for (Pair<CFGNode,E> c2 : summary) {
										/* Setup a link to match the needed exit node.
										   We need to include both exit and entry 
										   so we don't mess up the matching */
										TraceLink lnkExit = new TraceLink(
											pred, null, cur); 
										
										TraceLink lnkEntry = new TraceLink(
											c2.fst, c2.snd, lnkExit);
										
										addWorkListEntry(worklist, lnkEntry, true);
									} /* for each pair in summar */
								} /* if (summary != null) */
							} /* if (pred.isExit()) */
						} /* if link matches.. */
					} /* for (Pair<E,E> cand : predLinks) */
				} /* for (CFGNode pred : preds) */
			} /* if curConf is not endConf */
		} /* while (!worklist.isEmpty()) */

		/* Ugh. Oops */
		System.out.println("No trace available...");
	}
	
	/**
	 Issues a warning with given string, occuring at given configuration
	 This constructs a trace is possible
	*/
	static void warning(String warn, E warnConf) {
		if (!issued.contains(warn)) {
			/* Lookup the classification ... */
			Classification cf = classifyWarnings.get(warn);
			String mark = "UNKNOWN";
			if (cf == null) {
				++unknownWarn;
				++total;
			}
			else {
				mark = cf.toString();
				if (cf == Classification.Bug) {
					++validWarn;
					++total;
				}
				else {
					++invalidWarn;
					++total;
				}
			}
			System.out.print(warn);
			
			if (hasBugDB)
				System.out.println("Database classifies the warning as:" + mark);
				
 			if (traceOutput != null && traceOutput != System.out) {
 				traceOutput.print(warn);
 				if (hasBugDB)
 					traceOutput.println("Database classifies the warning as:" + mark);
 			}
			
			issued.add(warn);

			/* Construct a trace */
			if (warnConf != null)
				outputTrace(warnConf);
			
			System.out.println();
		}
	}
	
	/**
	 Initializes the warning machinery and the classifications database
	*/
	static void init() {
		/* Initialize node->function back table we need for debug..*/
		for (final Function fun : Symtab.getFunctions()) {
			CFGWalker init = new CFGWalker() {
				public void visit(CFGStatNode node) { 
					nodeFunctions.put(node, fun);
				}
				
			public void visit(CFGSwitchNode node) { 
				nodeFunctions.put(node, fun);
			}
				
			};
			init.visitAll(fun.getCFG());
		}
		
		/* See whether we have classifications file for some warnings */
		try {
			BufferedReader in = new BufferedReader(new FileReader(bugdbFile));
			System.out.println("Reading classifications from " + 
					bugdbFile + " ...\n");

			String line;
			
			Classification currentClassification = null;
			String currentWarning = null;
			
			boolean inClassify = false;
			
			while ((line = in.readLine()) != null) {
				/* The format has lines with @classify=bug and @end, 
				 * comments with #. Inspired by the KJS lut format */
				if (line.startsWith("#") || line.trim().length() == 0) 
					continue;

				if (line.startsWith("@classify ")) {
					if (inClassify) {
						System.err.println("Nested classifications, aborting read");
						break;
					}
					
					inClassify         = true;
					currentWarning     = "";
					String val = line.substring(line.indexOf(' ') + 1);
					if (val.equals("BUG"))
						currentClassification = Classification.Bug;
					else if (val.equals("FALSE"))
						currentClassification = Classification.FalsePositive;
					else
						currentClassification = null;
				} else if (line.equals("@end")) {
					if (!inClassify) {
						System.err.println("@end without classify, aborting read");
						break;
					}
					
					classifyWarnings.put(currentWarning, currentClassification);
					inClassify = false;
				} else {
					if (!inClassify) {
						System.err.println("Invalid text at top-level, aborting read");
						break;
					}
					
					currentWarning = currentWarning + line + "\n";
				}
			}
		} catch (FileNotFoundException e) {
			hasBugDB = false;
		} catch (IOException e) {
			System.err.println("[LC] Error while trying to read classifications:" + e);
		}
		
		/* Create trace output file */
		try {
			traceOutput = new PrintStream(new FileOutputStream(traceFile));
		} catch (IOException e) {
			System.err.println("Unable to open '" + traceFile + "', " +
					"traces will be printed to standard output");
			traceOutput = System.out;
		}
	}
	
	/* The next few fields contain info on the active probe used to 
	   construct the traces. Use the method below to set them...
	   
	   basically the sequence of calls to use is
	   newProbe();
	   probeOrigin();
	   probeOrigin();
	   probeStarted();
	   
	   ... and then may be warning
	*/
	
	
	/* Set of locations to which the probe corresponds. This can be multiple
	 * in the tail leak case.. */
	private static HashSet<String> probeStart = new HashSet<String>();
	private static CFGNode  probeStartNode;
	private static E        probeStartElement;
	private static IPAnalysisEngine<I,S,E,A> probeEng;
	
	static void probeStarted(IPAnalysisEngine<I,S,E,A> eng, CFGNode node, E element) {
		probeEng = eng;
		probeStartNode    = node;
		probeStartElement = element;
	}
	
	static void newProbe() {
		probeStart.clear();
	}
	
	static void probeOrigin(String msg) {
		probeStart.add(msg);
	}
	
	static void probeOrigin(CFGNode node) {
		probeStart.add(nodeInfo(node));
	}

	static String probeOriginNames() {
		StringBuilder out = new StringBuilder();
		for (String site : probeStart)
			out.append(site);
		return out.toString();
	}
}
